////////////////////////////////////////////////////////////////////////////////
//
// NormalTexture-Unit:
//
// Reads a normal texture from any UTPackage into a TBitmap-object,
// requieres a copy of the HeaderTables and the file-object as parameters.
//
// Includes: - TMipMap (contains unconverted texture and the palette)
//           - TNormalTexture (general properties and the converted texture)
//
// (c) 2000-2001, by Jesco Topp
////////////////////////////////////////////////////////////////////////////////
unit NormalTexture;

interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, extctrls, Math, JPEG,
  PCXctrl,
  PkgAux;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

type
  TMipMap = Class(TObject)
    WidthOffset    : DWORD; //(only if PkgVer >= 63)
    MipMapSize     : DWORD;
    MipMapData     : array [0..262143] of BYTE; //max 512*512 pixel
    Width          : DWORD;
    Height         : DWORD;
    BitsWidth      : BYTE;
    BitsHeight     : BYTE;
    constructor Create;
    procedure Free;
    destructor Destroy; override;
end;

//TNormalTexture
type
  TNormalTexture = Class(TObject)
    TextureName      : string;
    TexturePackage   : string;
    DetailTexture    : string;
    Palette          : string; //Name of used palette
    AnimNext         : string;
    MaxFrameRate     : single;
    HasMaxFramerate  : Boolean;
    MinFrameRate     : single;
    HasMinFramerate  : Boolean;
    UsedPalette      : TPalette;
    UBits            : BYTE;
    VBits            : BYTE;
    USize            : DWORD;
    VSize            : DWORD;
    UClamp           : DWORD;
    VClamp           : DWORD;
    bMasked          : Boolean;
    SerialSize       : DWORD;
    MipMapCount      : BYTE;
    MipMap           : TMipMap; //contains the actual Data and some properties
    B                : TBitmap; //used to create output
    MaxPreviewSize   : integer; //max. width or height of preview
    HasToBeResized   : Boolean; //if last available mipmap is larger than MaxPreviewSize, the mipmap has to be resized
    IsAnimated       : Boolean;
    constructor Create;
    procedure   Free;
    destructor  Destroy; override;
  public
    procedure Load(const NameTable : TNameTable; ImportTable : TImportTable; ExportTable : TExportTable; PkgVer : integer; Name : string; MaximumPreviewSize : integer; var F : File);
    procedure ReadGenericProperties(const NameTable : TNameTable; ImportTable : TImportTable; ExportTable : TExportTable; var Output : TStrings; var F : File);
    procedure Save(var NameTable : TNameTable; ImportTable : TImportTable; ExportTable : TExportTable; AddGroup : TAddGroup; Index : integer; var F : File);
  private
    Offset           : DWORD;
    function ReadUnknownProperty(const NameTable : TNameTable; ImportTable : TImportTable; ExportTable : TExportTable; Index : DWORD; var F : File) : string;
    procedure CreateBitmap;  //Creates the ouput-bitmap
end;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

implementation

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//TMipMap

constructor TMipMap.Create;
begin
end;

procedure TMipMap.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TMipMap.Destroy;
begin
  inherited Destroy;
end;

////////////////////////////////////////////////////////////////////////////////
//TNormalTexture

constructor TNormalTexture.Create;
begin
  MipMap   := TMipMap.Create;
  UsedPalette := TPalette.Create;
  B := TBitmap.Create;
  B.PixelFormat := pf8bit;
  IsAnimated := FALSE;
end;

procedure TNormalTexture.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TNormalTexture.Destroy;
begin
  MipMap.Free;
  UsedPalette.Free;
  B.Free;

  inherited Destroy;
end;

procedure TNormalTexture.Load;
var
  PropertyName : DWORD;
  TypeByte     : BYTE;
  IsUnknown    : Boolean;
  MipMapRead   : Boolean;
  MipMapsPassed: integer;
  //temp-stuff
  tempBYTE     : BYTE;
  tempDWORD    : DWORD;
  i            : integer;
begin
  // copy parameter to object inside TNormalTexture
  MaxPreviewSize := MaximumPreviewSize;

  //Get basic infos of texture (i.e. offset, SerialSize and actual name)
  for i:=0 to ExportTable.Count-1 do
    if (ExportTable.ExportClass.Strings[i] = 'Texture') then
      if (Name = ExportTable.ObjectName.Strings[i]) then
      begin
      Offset := StrToInt(ExportTable.SerialOffset.Strings[i]);
      SerialSize := StrToInt(ExportTable.SerialSize.Strings[i]);
      TextureName := ExportTable.ObjectName.Strings[i];
      TexturePackage := ExportTable.ExportPackage.Strings[i];

      break;
    end;

  ///////////////////////////////////////////////////////////
  //Decode properties
  IsUnknown := FALSE;
  HasMaxFrameRate := FALSE;
  HasMinFrameRate := FALSE;

  Seek(F, Offset);
  repeat
     // if property from last run was unknown then read it generic
     if ( IsUnknown = TRUE ) then
       ReadUnknownProperty(NameTable, ImportTable, ExportTable, PropertyName, F);

     // so far the new property is unknown
     IsUnknown := TRUE;

     PropertyName := ReadINDEX(F);

     //AnimNext
     if (NameTable.ObjectName[PropertyName] = 'AnimNext') then
     begin
       BlockRead(F, TypeByte, 1); //Should be 0x05 or 0x15 (ObjRef)
       AnimNext := GetObjectReference(NameTable, ImportTable, ExportTable, ReadINDEX(F));

       IsAnimated := TRUE;
       IsUnknown := FALSE;
     end;
     //MaxFrameRate
     if (NameTable.ObjectName[PropertyName] = 'MaxFrameRate') then
     begin
       BlockRead(F, TypeByte, 1); //Should be 0x24 (Float)
       BlockRead(F, MaxFrameRate, 4);
       HasMaxFrameRate := TRUE;

       IsAnimated := TRUE;
       IsUnknown := FALSE;
     end;
     //MinFrameRate
     if (NameTable.ObjectName[PropertyName] = 'MinFrameRate') then
     begin
       BlockRead(F, TypeByte, 1); //Should be 0x24 (Float)
       BlockRead(F, MinFrameRate, 4);
       HasMinFrameRate := TRUE;

       IsAnimated := TRUE;
       IsUnknown := FALSE;
     end;

     //bMasked
     if (NameTable.ObjectName.Strings[PropertyName] = 'bMasked') then
     begin
       BlockRead(F, TypeByte, 1); //0x53 = FALSE, 0xD3 = TRUE
       if ( TypeByte = $53 ) then bMasked := FALSE else bMasked := TRUE;
       BlockRead(F, tempByte, 1); //always 0x00
       IsUnknown := FALSE;
     end;

     //DetailTexture
     if (NameTable.ObjectName.Strings[PropertyName] = 'DetailTexture') then
     begin
       BlockRead(F, TypeByte, 1); //Should be 0x05 (=ObjectRef)
       DetailTexture := GetObjectReference(NameTable, ImportTable, ExportTable, ReadINDEX(F));
       IsUnknown := FALSE;
     end;

     //Palette
     if (NameTable.ObjectName.Strings[PropertyName] = 'Palette') then
     begin
       BlockRead(F, TypeByte, 1); //Should be 0x05 or 0x15 (=ObjectReference)
       Palette := GetObjectReference(NameTable, ImportTable, ExportTable, ReadINDEX(F));
       IsUnknown := FALSE;
     end;

     //UBits
     if (NameTable.ObjectName.Strings[PropertyName] = 'UBits') then
     begin
       BlockRead(F, TypeByte, 1); //Should be 0x01 (BYTE value follows)
       BlockRead(F, UBits, 1);
       IsUnknown := FALSE;
     end;
     //VBits
     if (NameTable.ObjectName.Strings[PropertyName] = 'VBits') then
     begin
       BlockRead(F, TypeByte, 1); //Should be 0x01 (BYTE value follows)
       BlockRead(F, VBits, 1);
       IsUnknown := FALSE;
     end;

     //USize
     if (NameTable.ObjectName.Strings[PropertyName] = 'USize') then
     begin
       BlockRead(F, TypeByte, 1); //Should be 0x22 (DWORD-Integer value follows)
       BlockRead(F, USize, 4);
       IsUnknown := FALSE;
     end;
     //VSize
     if (NameTable.ObjectName.Strings[PropertyName] = 'VSize') then
     begin
       BlockRead(F, TypeByte, 1); //Should 0x22 (DWORD-Integer value follows)
       BlockRead(F, VSize, 4);
       IsUnknown := FALSE;
     end;

     //UClamp
     if (NameTable.ObjectName.Strings[PropertyName] = 'UClamp') then
     begin
       BlockRead(F, TypeByte, 1); //Should be 0x22 (DWORD-Integer value follows)
       BlockRead(F, UClamp, 4);
       IsUnknown := FALSE;
     end;
     //VClamp
     if (NameTable.ObjectName.Strings[PropertyName] = 'VClamp') then
     begin
       BlockRead(F, TypeByte, 1); //Should be 0x22 (DWORD-Integer follows)
       BlockRead(F, VClamp, 4);
       IsUnknown := FALSE;
     end;
  until (NameTable.ObjectName.Strings[PropertyName] = 'None');

  ///////////////////////////////////////////////////////////
  //Read appropiate mipmap (i.e. max 128 pixels width/height)
  MipMapRead := FALSE;                         //No mipmap read yet
  MipMapsPassed := 0;                          //no mipmaps passed yet
  BlockRead(F, MipMapCount, 1);        //MipMap-Count
  repeat
    if (USize > VSize) then  //x >= y
    begin
      if (PkgVer >= 63) then                      //PackageVersion 63 or higher
      begin
        BlockRead(F, MipMap.WidthOffset, 4);
        MipMap.MipMapSize := ReadINDEX(F);

        Seek(F, MipMap.WidthOffset);          //Skip MipMapData

        BlockRead(F, MipMap.Width, 4);
        BlockRead(F, MipMap.Height, 4);
        BlockRead(F, MipMap.BitsWidth, 1);
        BlockRead(F, MipMap.BitsWidth, 1);

        if (MipMap.Width <= MaxPreviewSize) then
        begin
          Seek(F, MipMap.WidthOffset-MipMap.MipMapSize);
          BlockRead(F, MipMap.MipMapData, MipMap.MipMapSize);
          Seek(F, FilePos(F)+10);
          MipMapRead := TRUE;                        //MipMap is now read
          HasToBeResized := FALSE;
        end
        else
          if (MipMapsPassed = MipMapCount-1) then
          begin
            Seek(F, MipMap.WidthOffset-MipMap.MipMapSize);
            BlockRead(F, MipMap.MipMapData, MipMap.MipMapSize);
            Seek(F, FilePos(F)+10);
            MipMapRead := TRUE;                        //MipMap is now read
            HasToBeResized := TRUE;                    //mipmap is larger than required one
          end;
      end
      else                                           //PackageVersion lower than 63
      begin
        MipMap.MipMapSize := ReadINDEX(F);

        MipMap.WidthOffset := FilePos(F)+MipMap.MipMapSize; //generate missing offset
        Seek(F, FilePos(F)+MipMap.MipMapSize);             //Skip MipMapData

        BlockRead(F, MipMap.Width, 4);
        BlockRead(F, MipMap.Height, 4);
        BlockRead(F, MipMap.BitsWidth, 1);
        BlockRead(F, MipMap.BitsWidth, 1);

        if (MipMap.Width <= MaxPreviewSize) then
        begin
          Seek(F, FilePos(F)-MipMap.MipMapSize-10);
          BlockRead(F, MipMap.MipMapData, MipMap.MipMapSize);
          Seek(F, FilePos(F)+10);
          MipMapRead := TRUE;                        //MipMap is now read
          HasToBeResized := FALSE;
        end
        else
          if (MipMapsPassed = MipMapCount-1) then
          begin
            Seek(F, MipMap.WidthOffset-MipMap.MipMapSize);
            BlockRead(F, MipMap.MipMapData, MipMap.MipMapSize);
            Seek(F, FilePos(F)+10);
            MipMapRead := TRUE;                        //MipMap is now read
            HasToBeResized := TRUE;                    //mipmap is larger than required one
          end;
      end;
    end
    else                                                    //x < y
    begin
      if (PkgVer >= 63) then                                //PackageVersion 63 higher
      begin
        BlockRead(F, MipMap.WidthOffset, 4);
        MipMap.MipMapSize := ReadINDEX(F);

        Seek(F, MipMap.WidthOffset);          //Skip MipMapData

        BlockRead(F, MipMap.Width, 4);
        BlockRead(F, MipMap.Height, 4);
        BlockRead(F, MipMap.BitsWidth, 1);
        BlockRead(F, MipMap.BitsWidth, 1);

        if (MipMap.Height <= MaxPreviewSize) then
        begin
          Seek(F, MipMap.WidthOffset-MipMap.MipMapSize);
          BlockRead(F, MipMap.MipMapData, MipMap.MipMapSize);
          Seek(F, FilePos(F)+10);
          MipMapRead := TRUE;                        //MipMap is now read
          HasToBeResized := FALSE;
        end
        else
          if (MipMapsPassed = MipMapCount-1) then
          begin
            Seek(F, MipMap.WidthOffset-MipMap.MipMapSize);
            BlockRead(F, MipMap.MipMapData, MipMap.MipMapSize);
            Seek(F, FilePos(F)+10);
            MipMapRead := TRUE;                        //MipMap is now read
            HasToBeResized := TRUE;                    //mipmap is larger than required one
          end;
      end                                                      //PackageVersion lower than 63
      else
      begin
        MipMap.MipMapSize := ReadINDEX(F);

        MipMap.WidthOffset := FilePos(F)+MipMap.MipMapSize; //generate missing offset
        Seek(F, FilePos(F)+MipMap.MipMapSize);              //Skip MipMapData

        BlockRead(F, MipMap.Width, 4);
        BlockRead(F, MipMap.Height, 4);
        BlockRead(F, MipMap.BitsWidth, 1);
        BlockRead(F, MipMap.BitsWidth, 1);

        if (MipMap.Height <= MaxPreviewSize) then
        begin
          Seek(F, FilePos(F)-MipMap.MipMapSize-10);
          BlockRead(F, MipMap.MipMapData, MipMap.MipMapSize);
          Seek(F, FilePos(F)+10);
          MipMapRead := TRUE;                        //MipMap is now read
          HasToBeResized := FALSE;
        end
        else
          if (MipMapsPassed = MipMapCount-1) then
          begin
            Seek(F, MipMap.WidthOffset-MipMap.MipMapSize);
            BlockRead(F, MipMap.MipMapData, MipMap.MipMapSize);
            Seek(F, FilePos(F)+10);
            MipMapRead := TRUE;                        //MipMap is now read
            HasToBeResized := TRUE;                    //mipmap is larger than required one
          end;
      end;
    end;
    MipMapsPassed := MipMapsPassed + 1;
  until (MipMapRead = TRUE);

  ///////////////////////////////////////////////////////////
  //Read palette
  UsedPalette.Load(NameTable, ImportTable, ExportTable, F, Palette, TexturePackage);

  //Create bitmap
  CreateBitmap;
end;

procedure TNormalTexture.CreateBitmap;
var
  y,x          : integer;
  Row          : PByteArray;
  tempBitmap   : TBitmap;
  StretchRect  : TRect;
begin
  //Initialize everything
  B.Free;
  B := TBitmap.Create;
  B.Height := MipMap.Height;
  B.Width  := MipMap.Width;
  B.PixelFormat := pf8bit;

  //Assign Palette
  if (UsedPalette.WinPalette <> 0) then
    B.Palette := UsedPalette.WinPalette;
  //Create bitmap
  for y:=0 to MipMap.Height-1 do
  begin
    Row := B.ScanLine[y];
    for x:=0 to MipMap.Width-1 do
      Row[x] := MipMap.MipMapData[y*MipMap.Width+x];
  end;

  // Resize bitmap if larger than allowed
  if ( HasToBeResized = TRUE ) then
  begin
    tempBitmap := TBitmap.Create;
    tempBitmap.PixelFormat := pf8bit;

    if ( B.Width > B.Height ) then
    begin
      tempBitmap.Width  := MaxPreviewSize;
      tempBitmap.Height := MaxPreviewSize * ( MaxPreviewSize div B.Width );
    end;
    if ( B.Height > B.Width ) then
    begin
      tempBitmap.Height  := MaxPreviewSize;
      tempBitmap.Width := MaxPreviewSize * ( MaxPreviewSize div B.Height );
    end;
    if ( B.Width = B.Height ) then
    begin
      tempBitmap.Height  := MaxPreviewSize;
      tempBitmap.Width   := MaxPreviewSize;
    end;

    StretchRect.Top    := 0;
    StretchRect.Left   := 0;
    StretchRect.Right  := tempBitmap.Width-1;
    StretchRect.Bottom := tempBitmap.Height-1;
    
    //tempBitmap.Assign(B);
    tempBitmap.Canvas.StretchDraw( StretchRect, B);

    B.Assign(tempBitmap);
    tempBitmap.Free;
  end;
end;

// Read generic properties, used to display the enhanced properties in "Texture Details"
procedure TNormalTexture.ReadGenericProperties;
var
  Line            : string;
  Index           : integer;
begin
  Output.Add(' ');

  if ( IsAnimated = TRUE ) then
  begin
    Output.Add('Additional Properties of Frame ' + TextureName);
    Output.Add('---------------------------------------------');
    Output.Add(' ');
  end;

  Seek(F, Offset);

  Index := ReadINDEX(F);            // Read INDEX into NameTable for first loop

  while ( NameTable.ObjectName.Strings[Index] <> 'None' ) do
  begin
    Line := ReadUnknownProperty(NameTable, ImportTable, ExportTable, Index, F);

    // Add line to output StringList
    Output.Add(Line);

    // Read INDEX into NameTable for next pass of this loop
    Index := ReadINDEX(F);
    // Check for additional break-conditions (like unknown properties)
    if ( Index > NameTable.Count - 1 ) then
    begin
      Output.Add('Unknown property. Scan aborted.');
      break;
    end;
    if ( Index < 0 ) then
    begin
      Output.Add('Unknown property. Scan aborted.');
      break;
    end;
  end;
end;

function TNormalTexture.ReadUnknownProperty;
var
  Line            : string;
  TypeBYTE        : BYTE;
  ValueBYTE       : BYTE;
  ValueDWORD      : DWORD;
  ValueINDEX      : DWORD;
  ValueFloat      : single;
  ValueString     : string;
  ArrayIndex      : BYTE;
  StructINDEX     : DWORD;
begin
  // Reads one property
    Line  := NameTable.ObjectName.Strings[Index];

    BlockRead(F, TypeBYTE, 1);

    case TypeBYTE of
    $01 : // BYTE
      begin
        BlockRead(F, ValueBYTE, 1);
        Line := Line + ' (Byte) = ' + IntToStr(ValueBYTE);
      end;
    $81 : // BYTE + array
      begin
        BlockRead(F, ArrayIndex, 1);
        BlockRead(F, ValueBYTE, 1);
        Line := Line + '[' + IntToStr(ArrayINDEX) + '] (Byte) = ' + IntToStr(ValueBYTE);
      end;
    $05, $15 : // ObjRef
      begin
        ValueINDEX := ReadINDEX(F);
        Line := Line + ' (ObjRef) = ' + GetObjectReference(NameTable, ImportTable, ExportTable, ValueINDEX);
      end;
    $85, $95 : // ObjRef + array
      begin
        BlockRead(F, ArrayIndex, 1);
        ValueINDEX := ReadINDEX(F);
        Line := Line + '[' + IntToStr(ArrayIndex) + '] (ObjRef) = ' + GetObjectReference(NameTable, ImportTable, ExportTable, ValueINDEX);
      end;
    $55 : //ObjRef + unnown INDEX
      begin
        ValueINDEX := ReadINDEX(F);
        Line := Line + ' (ObjRef) = ' + IntToStr(ValueIndex);
        ValueINDEX := ReadINDEX(F);
        Line := Line + ', ' + GetObjectReference(NameTable, ImportTable, ExportTable, ValueINDEX);
      end;
    $D5 : //ObjRef + unnown INDEX + array
      begin
        Blockread(F, ArrayIndex, 1);
        ValueINDEX := ReadINDEX(F);
        Line := Line + '[' + IntToStr(ArrayIndex) + '] (ObjRef) = ' + IntToStr(ValueIndex);
        ValueINDEX := ReadINDEX(F);
        Line := Line + ', ' + GetObjectReference(NameTable, ImportTable, ExportTable, ValueINDEX);
      end;
    $06, $16 : // Index into NameTable
      begin
        ValueINDEX := ReadINDEX(F);
        Line := Line + ' (Name) = ' + IntToStr(ValueINDEX);
      end;
    $86, $96 : // Index into NameTable
      begin
        BlockRead(F, ArrayIndex, 1);
        ValueINDEX := ReadINDEX(F);
        Line := Line + '[' + IntToStr(ArrayIndex) + '] (Name) = ' + IntToStr(ValueINDEX);
      end;
    $08 : // ADrop
      begin
        BlockRead(F, ValueDWORD, 4);
        BlockRead(F, ValueBYTE, 1);
        Line := Line + ' (ADrop) = [' + IntToStr(ValueDWORD) + ', x = ' + IntToStr(ValueBYTE);
        BlockRead(F, ValueBYTE, 1);
        BlockRead(F, ValueDWORD, 4);
        Line := Line + ', y = ' + IntToStr(ValueBYTE) + ', ' + IntToStr(ValueDWORD) + ']';
      end;
    $88 : // ADrop + array
      begin
        BlockRead(F, ArrayIndex, 1);
        BlockRead(F, ValueDWORD, 4);
        BlockRead(F, ValueBYTE, 1);
        Line := Line + '[' + IntToStr(ArrayIndex) + '] (ADrop) = [' + IntToStr(ValueDWORD) + ', x = ' + IntToStr(ValueBYTE);
        BlockRead(F, ValueBYTE, 1);
        BlockRead(F, ValueDWORD, 4);
        Line := Line + ', y = ' + IntToStr(ValueBYTE) + ', ' + IntToStr(ValueDWORD) + ']';
      end;
    $0D : // Localized string
      begin
        ValueINDEX := ReadINDEX(F);
        BlockRead(F, ValueString, ValueINDEX);
        Line := Line + ' (String) = ' + ValueString;
      end;
    $8D : // Localized string + array
      begin
        BlockRead(F, ArrayIndex, 1);
        ValueINDEX := ReadINDEX(F);
        BlockRead(F, ValueString, ValueINDEX);
        Line := Line + '[' + IntToStr(ArrayIndex) + '] (String) = ' + ValueString;
      end;
    $22 : // Integer
      begin
        BlockRead(F, ValueDWORD, 4);
        Line := Line + ' (Integer) = ' + IntToStr(ValueDWORD);
      end;
    $A2 : // Integer + array
      begin
        BlockRead(F, ArrayIndex, 1);
        BlockRead(F, ValueDWORD, 4);
        Line := Line + '[' + IntToStr(ArrayIndex) + '] (Integer) = ' + IntToStr(ValueDWORD);
      end;
    $24 : // Floating point
      begin
        BlockRead(F, ValueFloat, 4);
        Line := Line + ' (Float) = ' + FloatToStr(ValueFloat);     //buggy
      end;
    $A4 : // Floating point + array
      begin
        BlockRead(F, ArrayIndex, 1);
        BlockRead(F, ValueFloat, 4);
        Line := Line + '[' + IntToStr(ArrayIndex) + '] (Float) = ' + FloatToStr(ValueFloat);  //buggy
      end;
    $3D, $4D : // Text
      begin
        BlockRead(F, ValueBYTE, 1);
        BlockRead(F, ValueString, ValueBYTE);
        Line := Line + ' (Text) = ' + ValueString;
      end;
    $BD, $CD : // Text + array
      begin
        BlockRead(F, ArrayIndex, 1);
        BlockRead(F, ValueBYTE, 1);
        BlockRead(F, ValueString, ValueBYTE);
        Line := Line + '[' + IntToStr(ArrayIndex) + '] (Text) = ' + ValueString;
      end;
    $5D : // Text
      begin
        BlockRead(F, ValueBYTE, 1);
        ValueINDEX := ReadINDEX(F);
        BlockRead(F, ValueString, ValueINDEX);
        Line := Line + ' (Text) = ' + ValueString;
      end;
    $DD : // Text + array
      begin
        BlockRead(F, ArrayIndex, 1);
        BlockRead(F, ValueBYTE, 1);
        ValueINDEX := ReadINDEX(F);
        BlockRead(F, ValueString, ValueINDEX);
        Line := Line + '[' + IntToStr(ArrayIndex) + '] (Text) = ' + ValueString;
      end;
    $6D : // Text
      begin
        BlockRead(F, ValueBYTE, 1);
        BlockRead(F, ValueBYTE, 1);
        ValueINDEX := ReadINDEX(F);
        BlockRead(F, ValueString, ValueINDEX);
        Line := Line + ' (Text) = ' + ValueString;
      end;
    $ED : // Text + array
      begin
        BlockRead(F, ArrayIndex, 1);
        BlockRead(F, ValueBYTE, 1);
        BlockRead(F, ValueBYTE, 1);
        ValueINDEX := ReadINDEX(F);
        BlockRead(F, ValueString, ValueINDEX);
        Line := Line + '[' + IntToStr(ArrayIndex) + '] (Text) = ' + ValueString;
      end;
    $53 : // bool = FALSE
      begin
        BlockRead(F, ValueBYTE, 1);
        Line := Line + ' (Boolean) = FALSE';
      end;
    $D3 : // bool = TRUE
      begin
        BlockRead(F, ValueBYTE, 1);
        Line := Line + ' (Boolean) = TRUE';
      end;
    $2A, $3A : // struct
      begin
        StructINDEX := ReadINDEX(F);

        // Color struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Color' ) then
        begin
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ' (Color) = [R=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';G=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';B=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';A=' + IntToStr(ValueBYTE) + ']';
        end;
        // Vector struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Vector' ) then
        begin
          BlockRead(F, ValueFloat, 4);
          Line := Line + ' (Vector) = [x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; y=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; x=' + FloatToStr(ValueFloat) + ']';
        end;
        // Rotator struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Rotator' ) then
        begin
          BlockRead(F, ValueDWORD, 4);
          Line := Line + ' (Rotator) = [pitch=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; yaw=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; roll=' + IntToStr(ValueDWORD) + ']';
        end;
        // Scale struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Scale' ) then
        begin
          BlockRead(F, ValueFloat, 4);
          Line := Line + ' (Scale) = [x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; y=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; sheerrate=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + '; sheeraxis=' + IntToStr(ValueBYTE) + ']';
        end;
        // PointRegion
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'PointRegion' ) then
        begin
          ValueINDEX := ReadINDEX(F);
          Line := Line + ' (PointRegion) = [zone=' + GetObjectReference(NameTable, ImportTable, ExportTable, ValueINDEX);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; ileaf=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + '; zonenumber=' + IntToStr(ValueBYTE) + ']';
        end;
      end;
    $AA, $BA : //struct + array
      begin
        StructINDEX := ReadINDEX(F);
        BlockRead(F, ArrayIndex, 1);

        // Color struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Color' ) then
        begin
          BlockRead(F, ValueBYTE, 1);
          Line := Line + '[' + IntToStr(ArrayIndex) + '] (Color) = [R=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';G=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';B=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';A=' + IntToStr(ValueBYTE) + ']';
        end;
        // Vector struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Vector' ) then
        begin
          BlockRead(F, ValueFloat, 4);
          Line := Line + '[' + IntToStr(ArrayIndex) + '] (Vector) = [x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; y=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; x=' + FloatToStr(ValueFloat) + ']';
        end;
        // Rotator struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Rotator' ) then
        begin
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '[' + IntToStr(ArrayIndex) + '] (Rotator) = [pitch=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; yaw=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; roll=' + IntToStr(ValueDWORD) + ']';
        end;
        // Scale struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Scale' ) then
        begin
          BlockRead(F, ValueFloat, 4);
          Line := Line + '[' + IntToStr(ArrayIndex) + '] (Scale) = [x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; y=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; sheerrate=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + '; sheeraxis=' + IntToStr(ValueBYTE) + ']';
        end;
        // PointRegion
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'PointRegion' ) then
        begin
          ValueINDEX := ReadINDEX(F);
          Line := Line + '[' + IntToStr(ArrayIndex) + '] (PointRegion) = [zone=' + GetObjectReference(NameTable, ImportTable, ExportTable, ValueINDEX);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; ileaf=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + '; zonenumber=' + IntToStr(ValueBYTE) + ']';
        end;
      end;
    $5A : // struct
      begin
        StructINDEX := ReadINDEX(F);
        ValueINDEX := ReadINDEX(F);

        // Color struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Color' ) then
        begin
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ' (Color) = [R=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';G=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';B=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';A=' + IntToStr(ValueBYTE) + ']';
        end;
        // Vector struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Vector' ) then
        begin
          BlockRead(F, ValueFloat, 4);
          Line := Line + ' (Vector) = [x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; y=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; x=' + FloatToStr(ValueFloat) + ']';
        end;
        // Rotator struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Rotator' ) then
        begin
          BlockRead(F, ValueDWORD, 4);
          Line := Line + ' (Rotator) = [pitch=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; yaw=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; roll=' + IntToStr(ValueDWORD) + ']';
        end;
        // Scale struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Scale' ) then
        begin
          BlockRead(F, ValueFloat, 4);
          Line := Line + ' (Scale) = [x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; y=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; sheerrate=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + '; sheeraxis=' + IntToStr(ValueBYTE) + ']';
        end;
        // PointRegion
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'PointRegion' ) then
        begin
          ValueINDEX := ReadINDEX(F);
          Line := Line + ' (PointRegion) = [zone=' + GetObjectReference(NameTable, ImportTable, ExportTable, ValueINDEX);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; ileaf=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + '; zonenumber=' + IntToStr(ValueBYTE) + ']';
        end;
      end;
    $DA : //struct + array
      begin
        StructINDEX := ReadINDEX(F);
        ValueINDEX := ReadINDEX(F);
        BlockRead(F, ArrayIndex, 1);

        // Color struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Color' ) then
        begin
          BlockRead(F, ValueBYTE, 1);
          Line := Line + '[' + IntToStr(ArrayIndex) + '] (Color) = [R=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';G=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';B=' + IntToStr(ValueBYTE);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + ';A=' + IntToStr(ValueBYTE) + ']';
        end;
        // Vector struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Vector' ) then
        begin
          BlockRead(F, ValueFloat, 4);
          Line := Line + '[' + IntToStr(ArrayIndex) + '] (Vector) = [x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; y=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; x=' + FloatToStr(ValueFloat) + ']';
        end;
        // Rotator struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Rotator' ) then
        begin
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '[' + IntToStr(ArrayIndex) + '] (Rotator) = [pitch=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; yaw=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; roll=' + IntToStr(ValueDWORD) + ']';
        end;
        // Scale struct
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'Scale' ) then
        begin
          BlockRead(F, ValueFloat, 4);
          Line := Line + '[' + IntToStr(ArrayIndex) + '] (Scale) = [x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; y=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueFloat, 4);
          Line := Line + '; x=' + FloatToStr(ValueFloat);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; sheerrate=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + '; sheeraxis=' + IntToStr(ValueBYTE) + ']';
        end;
        // PointRegion
        if ( NameTable.ObjectName.Strings[StructINDEX] = 'PointRegion' ) then
        begin
          ValueINDEX := ReadINDEX(F);
          Line := Line + '[' + IntToStr(ArrayIndex) + '] (PointRegion) = [zone=' + GetObjectReference(NameTable, ImportTable, ExportTable, ValueINDEX);
          BlockRead(F, ValueDWORD, 4);
          Line := Line + '; ileaf=' + IntToStr(ValueDWORD);
          BlockRead(F, ValueBYTE, 1);
          Line := Line + '; zonenumber=' + IntToStr(ValueBYTE) + ']';
        end;
      end;
    end;

  // pass Line back
  Result := Line;
end;


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

//Save texture
procedure TNormalTexture.Save;
var
  PCXBitmap        : TPCXBitmap;                //original .PCX file
  ResizeBitmap     : TBitmap;                   //used to resize bitmap
  Color            : array [0..255] of TColor;  //stores palette-entries
  Data             : array [0..262143] of BYTE; //stores indices into palette
  ColorCount       : integer;                   //number of colors in palette
  MipWidth,
   MipHeight       : integer;                   //used to get the count of mipmaps
  Height, Width    : integer;                   //package-writer ...
  TypeBYTE         : BYTE;
  InvalidSize      : Boolean;
  tempColor        : TColor;
  tempDWORD        : DWORD;
  tempBYTE         : BYTE;
  tempOffset1      : DWORD;
  tempOffset2      : DWORD;                     //... package-writer
  x,y              : integer;                   //coordinates in bitmap
  i, counter, k    : integer;                   //counter
  //Additional Properties
    MipZero          : TUTColor;              //(obselote)
    MaxColor         : TUTColor;              //(obselote)
    InternalTime     : array [0..1] of DWORD; //(obselote)
begin
  //----------------------
  // Load texture into the memory and check if the format is valid

  //Load texture from file and convert it to .BMP
  PCXBitmap := TPCXBitmap.Create;
  PCXBitmap.LoadFromFile(AddGroup.FileName.Strings[Index]);
  B.Assign(PCXBitmap);
  PCXBitmap.Destroy;

  //Check if texture-size is power of two (and smaller than 512*512)
  if ( B.Width > 512 ) then Raise TIllegalPackageOperationException.Create('Texture ' + AddGroup.TextureName[Index] + ' is larger than 512*512. Maximum supported size is 512*512.');
  if ( B.Height > 512 ) then Raise TIllegalPackageOperationException.Create('Texture ' + AddGroup.TextureName[Index] + ' is larger than 512*512. Maximum supported size is 512*512.');

  //check width
  case B.Width of
    1   : begin InvalidSize := FALSE; MipWidth := 1; end;
    2   : begin InvalidSize := FALSE; MipWidth := 2; end;
    4   : begin InvalidSize := FALSE; MipWidth := 3; end;
    8   : begin InvalidSize := FALSE; MipWidth := 4; end;
    16  : begin InvalidSize := FALSE; MipWidth := 5; end;
    32  : begin InvalidSize := FALSE; MipWidth := 6; end;
    64  : begin InvalidSize := FALSE; MipWidth := 7; end;
    128 : begin InvalidSize := FALSE; MipWidth := 8; end;
    256 : begin InvalidSize := FALSE; MipWidth := 9; end;
    512 : begin InvalidSize := FALSE; MipWidth := 10; end;
  else
    InvalidSize := TRUE;
  end;
  //If width is correct, check height
  if ( InvalidSize = FALSE ) then
    case B.Height of
    1   : begin InvalidSize := FALSE; MipHeight := 1; end;
    2   : begin InvalidSize := FALSE; MipHeight := 2; end;
    4   : begin InvalidSize := FALSE; MipHeight := 3; end;
    8   : begin InvalidSize := FALSE; MipHeight := 4; end;
    16  : begin InvalidSize := FALSE; MipHeight := 5; end;
    32  : begin InvalidSize := FALSE; MipHeight := 6; end;
    64  : begin InvalidSize := FALSE; MipHeight := 7; end;
    128 : begin InvalidSize := FALSE; MipHeight := 8; end;
    256 : begin InvalidSize := FALSE; MipHeight := 9; end;
    512 : begin InvalidSize := FALSE; MipHeight := 10; end;
    else
      InvalidSize := TRUE;
    end;
  //If width or height are incorrect, display exception
  if ( InvalidSize = TRUE ) then Raise TIllegalPackageOperationException.Create('Size of texture ' + AddGroup.TextureName[Index] + ' is not a power of two, e.g. 128*64 or 256*256.');


  //Get proper MipMapCount
  if ( AddGroup.HasMipMaps.Strings[Index] = 'TRUE' ) then
  begin
    if ( MipWidth >= MipHeight ) then MipMapCount := MipWidth-1 else MipMapCount := MipHeight-1;
  end
  else
    MipMapCount := 1;

  //----------------------
  // Create UTPalette and save it to disk & create the Table-entries for palette
  // Create first mipmap

  //////////////////////////////////////
  //Load palette and create first mipmap-data
  ColorCount := 0;
  for y:=0 to B.Height-1 do
    for x:=0 to B.Width-1 do
    begin
      tempColor := B.Canvas.Pixels[x,y];
      if ( ColorCount > 0 ) then
        for i:=0 to ColorCount-1 do
        begin
          if ( tempColor = Color[i] ) then      //Entry already exists
          begin
            Data[B.Width*y+x] := i;
            break;
          end
          else
            if ( i = ColorCount-1 ) then        //last color-entry and still no match found
            begin
              Color[i+1] := tempColor;
              Data[B.Width*y+x] := i+1;
              ColorCount := ColorCount + 1;
            end;
        end
      else
      begin
        Color[0] := tempColor;
        Data[0] := 0;
        ColorCount := ColorCount + 1;
      end;
    end;


    //////////////////////////////////////
    //Create UTPalette
    UsedPalette.PaletteSize := ColorCount;
    UsedPalette.PaletteName := AddGroup.TextureName[Index] + 'pal';
    Palette := UsedPalette.PaletteName;
    for i:=0 to UsedPalette.PaletteSize-1 do
    begin
      tempDWORD := Color[i];
      UsedPalette.Color[i].r := BYTE(tempDWORD);

      tempDWORD := Color[i];
      UsedPalette.Color[i].g := BYTE(tempDWORD shr 8);

      tempDWORD := Color[i];
      UsedPalette.Color[i].b := BYTE(tempDWORD shr 16);
    end;
    UsedPalette.PaletteOffset := FilePos(F);

    //Write palette to disk
    //Properties
    tempDWORD := 0;
    WriteINDEX(tempDWORD, F);  //Index to None in NameTable
    //Data
    tempDWORD := UsedPalette.PaletteSize;
    WriteINDEX(tempDWORD, F);
    for i:=0 to UsedPalette.PaletteSize-1 do
    begin
      BlockWrite(F, UsedPalette.Color[i].r, 1);
      BlockWrite(F, UsedPalette.Color[i].g, 1);
      BlockWrite(F, UsedPalette.Color[i].b, 1);
      tempBYTE := 0;
      BlockWrite(F, tempBYTE, 1);
    end;

    //Create Table-Entries (palette)
    ExportTable.ExportClass.Add(IntToStr(-2));
    ExportTable.Super.Add('0');
    ExportTable.ExportPackage.Add(IntToStr(AddGroup.GroupIndex+1));
     NameTable.ObjectName.Add(UsedPalette.PaletteName);
     NameTable.ObjectFlags.Add(IntToStr($00010000 or $00020000 or $00040000 or $00000010));
    ExportTable.ObjectName.Add(IntToStr(NameTable.ObjectName.Count-1));
    ExportTable.ObjectFlags.Add(IntToStr($00010000 or $00020000 or $00040000 or $00000004));
    ExportTable.SerialSize.Add(IntToStr(1 + SizeOfINDEX(UsedPalette.PaletteSize) + UsedPalette.PaletteSize*4));
    ExportTable.SerialOffset.Add(IntToStr(UsedPalette.PaletteOffset));

    //----------------------
    // Save texture to disk

    TextureName := AddGroup.TextureName[Index];
    Offset      := FilePos(F);

    //Properties//
    //Palette
    tempDWORD := 9; //points to Palette in NameTable
    WriteINDEX(tempDWORD, F);
    TypeBYTE := $05; //Type is ObjRef
    BlockWrite(F, TypeBYTE, 1);
    tempDWORD := ExportTable.ObjectName.Count; //points to last entry in ExportTable (Exactly: Count - 1 + 1; ObjRef)
    WriteINDEX(tempDWORD, F);

    //UBits
    tempDWORD := 5; //points to UBits in NameTable
    WriteINDEX(tempDWORD, F);
    TypeBYTE := $01; //Type is BYTE
    BlockWrite(F, TypeBYTE, 1);
    if ( B.Width = 1  ) then UBits := 0;
    if ( B.Width = 2  ) then UBits := 1;
    if ( B.Width = 4  ) then UBits := 2;
    if ( B.Width = 8  ) then UBits := 3;
    if ( B.Width = 16 ) then UBits := 4;
    if ( B.Width = 32 ) then UBits := 5;
    if ( B.Width = 64 ) then UBits := 6;
    if ( B.Width = 128) then UBits := 7;
    if ( B.Width = 256) then UBits := 8;
    if ( B.Width = 512) then UBits := 9;
    BlockWrite(F, UBits, 1);

    //VBits
    tempDWORD := 6; //points to UBits in NameTable
    WriteINDEX(tempDWORD, F);
    TypeBYTE := $01; //Type is BYTE
    BlockWrite(F, TypeBYTE, 1);
    if ( B.Height = 1  ) then VBits := 0;
    if ( B.Height = 2  ) then VBits := 1;
    if ( B.Height = 4  ) then VBits := 2;
    if ( B.Height = 8  ) then VBits := 3;
    if ( B.Height = 16 ) then VBits := 4;
    if ( B.Height = 32 ) then VBits := 5;
    if ( B.Height = 64 ) then VBits := 6;
    if ( B.Height = 128) then VBits := 7;
    if ( B.Height = 256) then VBits := 8;
    if ( B.Height = 512) then VBits := 9;
    BlockWrite(F, VBits, 1);

    //USize
    tempDWORD := 7; //points to USize in NameTable
    WriteINDEX(tempDWORD, F);
    TypeBYTE := $22;
    BlockWrite(F, TypeBYTE, 1);
    USize := B.Width;
    BlockWrite(F, USize, 4);

    //VSize
    tempDWORD := 8; //points to VSize in NameTable
    WriteINDEX(tempDWORD, F);
    TypeBYTE := $22;
    BlockWrite(F, TypeBYTE, 1);
    VSize := B.Height;
    BlockWrite(F, VSize, 4);

    //UClamp
    tempDWORD := 10; //points to UClamp in NameTable
    WriteINDEX(tempDWORD, F);
    TypeBYTE := $22;
    BlockWrite(F, TypeBYTE, 1);
    UClamp := B.Width;
    BlockWrite(F, UClamp, 4);

    //VClamp
    tempDWORD := 11; //points to VClamp in NameTable
    WriteINDEX(tempDWORD, F);
    TypeBYTE := $22;
    BlockWrite(F, TypeBYTE, 1);
    VClamp := B.Height;
    BlockWrite(F, VClamp, 4);

    //InternalTime
    tempDWORD := 1; //points to InternalTime
    WriteINDEX(tempDWORD, F);
    TypeBYTE := $22; //first array-entry
    BlockWrite(F, TypeBYTE, 1);
    InternalTime[0] := $05F36CD8;
    BlockWrite(F, InternalTime[0], 4);
    tempDWORD := 1; //points to InternalTime
    WriteINDEX(tempDWORD, F);
    TypeBYTE := $A2; //first array-entry
    BlockWrite(F, TypeBYTE, 1);
    tempDWORD := $01; //array-index
    WriteINDEX(tempDWORD, F);
    InternalTime[1] := $0032DD65;
    BlockWrite(F, InternalTime[1], 4);

    //bMasked (if checked)
    if ( AddGroup.bMasked.Strings[Index] = 'TRUE' ) then
    begin
      tempDWORD := 17; //points to bMasked in NameTable
      WriteINDEX(tempDWORD, F);
      TypeBYTE := $D3;
      BlockWrite(F, TypeBYTE, 1);
      tempBYTE := 0;
      BlockWrite(F, tempBYTE, 1);
    end;

    //None
    tempDWORD := 0; //points to None in NameTable
    WriteINDEX(tempDWORD, F);


    //Write additional properties//
    //MipMapCount
    BlockWrite(F, MipMapCount, 1);

    //width and height for first mipmap
    Width := B.Width;
    Height := B.Height;
    for counter:=0 to MipMapCount-1 do
    begin
      //Write MipMap-Properties and Data//
      tempOffset1 := FilePos(F);
      MipMap.WidthOffset := 0; //dummy
      BlockWrite(F, MipMap.WidthOffset, 4);
      MipMap.MipMapSize := Width*Height;
      WriteINDEX(MipMap.MipMapSize, F);
      for i:=0 to MipMap.MipMapSize-1 do
      begin
        MipMap.MipMapData[i] := Data[i];
        //BlockWrite(F, MipMap.MipMapData[i], 1);
      end;
      BlockWrite(F, MipMap.MipMapData, MipMap.MipMapSize);

      tempOffset2 := FilePos(F);
      MipMap.WidthOffset := FilePos(F);
      Seek(F, tempOffset1);
      BlockWrite(F, MipMap.WidthOffset, 4);
      Seek(F, tempOffset2);

      MipMap.Width := Width;
      BlockWrite(F, MipMap.Width, 4);
      MipMap.Height := Height;
      BlockWrite(F, MipMap.Height, 4);
      MipMap.BitsWidth := UBits - counter;
      BlockWrite(F, MipMap.BitsWidth, 1);
      MipMap.BitsHeight := VBits - counter;
      BlockWrite(F, MipMap.BitsHeight, 1);

      //Prepare data for next mipmap
      ResizeBitmap := TBitmap.Create;
      ResizeBitmap.Assign(B);      //needed for correct-palette
      //Get height and width of next mipmap (counter+1)
      case (VBits - counter - 1) of
        9 : ResizeBitmap.Height := 512;
        8 : ResizeBitmap.Height := 256;
        7 : ResizeBitmap.Height := 128;
        6 : ResizeBitmap.Height := 64;
        5 : ResizeBitmap.Height := 32;
        4 : ResizeBitmap.Height := 16;
        3 : ResizeBitmap.Height := 8;
        2 : ResizeBitmap.Height := 4;
        1 : ResizeBitmap.Height := 2;
        0 : begin ResizeBitmap.Height := 1; break; end;
      end;
      case (UBits - counter - 1) of
        9 : ResizeBitmap.Width := 512;
        8 : ResizeBitmap.Width := 256;
        7 : ResizeBitmap.Width := 128;
        6 : ResizeBitmap.Width := 64;
        5 : ResizeBitmap.Width := 32;
        4 : ResizeBitmap.Width := 16;
        3 : ResizeBitmap.Width := 8;
        2 : ResizeBitmap.Width := 4;
        1 : ResizeBitmap.Width := 2;
        0 : begin ResizeBitmap.Width := 1; break; end;
      end;
      Height := ResizeBitmap.Height;
      Width  := ResizeBitmap.Width;

      //Resize bitmap
      ResizeBitmap.Canvas.StretchDraw( Rect(0,0, Width, Height), B);
      //ResizeBitmap.SaveToFile('C:\TEMP\ResizeBMP'+ IntToStr(counter) + '.bmp');

      //get indices into palette
      for y:=0 to ResizeBitmap.Height-1 do
        for x:=0 to ResizeBitmap.Width-1 do
        begin
          tempColor := ResizeBitmap.Canvas.Pixels[x,y];
          for k:=0 to ColorCount-1 do
            if ( tempColor = Color[k] ) then
            begin
              Data[ResizeBitmap.Width*y+x] := k;
              break;
            end;
        end;

      ResizeBitmap.Free;
    end;

    SerialSize := FilePos(F) - Offset;

    //----------------------
    // Create Table-Entries
    ExportTable.ExportClass.Add(IntToStr(-3));
    ExportTable.Super.Add('0');
    ExportTable.ExportPackage.Add(IntToStr(AddGroup.GroupIndex+1));
     NameTable.ObjectName.Add(TextureName);
     NameTable.ObjectFlags.Add(IntToStr($00010000 or $00020000 or $00040000 or $00000010));
    ExportTable.ObjectName.Add(IntToStr(NameTable.ObjectName.Count-1));
    ExportTable.ObjectFlags.Add(IntToStr($00010000 or $00020000 or $00040000 or $00000004 or $00080000));
    ExportTable.SerialSize.Add(IntToStr(SerialSize));
    ExportTable.SerialOffset.Add(IntToStr(Offset));
end;

end.
