////////////////////////////////////////////////////////////////////////////////
//
// PackageFormat-Unit:
//
// Includes: - TUTPackage object (main object, used for all other operations)
//
// (c) 2000-2001, by Jesco Topp
////////////////////////////////////////////////////////////////////////////////
unit PackageFormat;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  StdCtrls, ExtCtrls, Forms, Spin, Dialogs,
  PkgAux,
  NormalTexture;

////////////////////////////////////////////////////////////////////////////////

//TGroup
type
  TGroup = Class(TObject)
    Name    : string;
    Entries : TStrings;
    constructor Create;
    procedure Free;
    destructor Destroy; override;
end;

////////////////////////////////////////////////////////////////////////////////
//Main-UTPackage-Object

type
  TUTPackage = Class(TObject)
    //Header-Data
    Signature      : DWORD;
    PkgVer         : DWORD;
    PackageFlags   : DWORD;
    //HeaderTables
    NameTable      : TNameTable;
    ImportTable    : TImportTable;
    ExportTable    : TExportTable;
    //List of internal packages
    Group          : array [0..127] of TGroup;
    AddGroup       : array [0..127] of TAddGroup;
    GroupCount     : integer;
    //Texture related stuff
    NormalTexture  : TNormalTexture; //normal texture (properties, data and palettes)
    TextureCount   : integer;
    //Misc stuff
    PackageSize    : longint;
    Name           : string;
    //Load/Save related stuff
    MODE           : DWORD;          //I/O-MODE : Read-only or Write-only
    //Functions
    constructor Create;
    procedure   Free;
    destructor  Destroy; override;
  public
    F              : File;
    //Load-functions
    function    Load(const FileName : string) : Boolean; //Load a package and its headers
    procedure   LoadNormalTexture(const Name : string; const MaxPreviewSize : integer); //Loads a normal texture (meaning no scripted ones like FireTexture)
    //Save functions
    procedure   Save(const FileName : string);  //Save a package with all of its content
  private
    IsPkgOpen      : Boolean;
    procedure CreateGroupList;           //Create a list of internal packages
end;
//////////////////////////////
//Constants

const
  PKG_READ  = 0;
  PKG_WRITE = 1;


implementation

uses
  Main;

////////////////////////////////////////////////////////////////////////////////

//TGroup
constructor TGroup.Create;
begin
  Entries := TStringList.Create;
end;

procedure TGroup.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TGroup.Destroy;
begin
  Entries.Free;

  inherited Destroy;
end;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//TUTPackage

////////////////////////////////////////////////////////////////////////////////
//Load

constructor TUTPackage.Create;
var
  i         : integer;
begin
  //Create all objects
  NameTable     := TNameTable.Create;
  ImportTable   := TImportTable.Create;
  ExportTable   := TExportTable.Create;
  NormalTexture := TNormalTexture.Create;
  for i:=0 to 127 do
    Group[i] := TGroup.Create;
  for i:=0 to 127 do
    AddGroup[i] := TAddGroup.Create;
  GroupCount := 0;

  //No file currently opened
  IsPkgOpen := FALSE;
  //Randomizer
  Randomize;
end;

procedure TUTPackage.Free;
begin
 if Self <> nil then Destroy;
end;

destructor TUTPackage.Destroy;
var
  i        : integer;
begin
  //Destroy everything
  NameTable.Free;
  ImportTable.Free;
  ExportTable.Free;
  NormalTexture.Free;
  for i:=0 to 127 do
    Group[i].Free;
  for i:=0 to 127 do
    AddGroup[i].Free;

  //Check if file is open or not
  if (IsPkgOpen = TRUE) then
    CloseFile(F);
end;

//Loads a package and its headers
function TUTPackage.Load;
var
  i      : integer;
  temp   : DWORD;
begin
  //Setup everything
  AssignFile(F, FileName);
  FileMode := 0;
  Reset(F,1);
  IsPkgOpen := TRUE;
  Name := FileName;

  //Get filesize
  PackageSize := FileSize(F);

  //Read Header and tables
  BlockRead(F, Signature, 4);
  if (Signature <> $9E2A83C1) then
    Result := FALSE
  else
  begin
    BlockRead(F, PkgVer, 4);      //Read PackageVersion
    BlockRead(F, PackageFlags, 4);//Read Package-Flags
    BlockRead(F, NameTable.Count, 4);   //Read NameCount & Offset
    BlockRead(F, NameTable.Offset, 4);
    BlockRead(F, ExportTable.Count, 4); //Read ExportCount & Offset
    BlockRead(F, ExportTable.Offset, 4);
    BlockRead(F, ImportTable.Count, 4); //Read ImportCount & Offset
    BlockRead(F, ImportTable.Offset, 4);

    //Read NameTable
    Seek(F, NameTable.Offset);
    for i := 0 to NameTable.Count - 1 do
    begin
      NameTable.ObjectName.Add(ReadNAME(F, PkgVer));
      BlockRead(F, temp, 4);
      NameTable.ObjectFlags.Add(IntToStr(temp));
    end;

    //Read ImportTable
    Seek(F, ImportTable.Offset);
    for i := 0 to ImportTable.Count - 1 do
    begin
      ImportTable.ClassPackage.Add(NameTable.ObjectName.Strings[ReadINDEX(F)]);
      ImportTable.ClassName.Add(NameTable.ObjectName.Strings[ReadINDEX(F)]);
      BlockRead(F, temp, 4); ImportTable.ImportPackage.Add(IntToStr(temp));
      ImportTable.ObjectName.Add(NameTable.ObjectName.Strings[ReadINDEX(F)]);
    end;

    //Read ExportTable
    Seek(F, ExportTable.Offset);
    for i := 0 to ExportTable.Count - 1 do
    begin
      temp := ReadINDEX(F);
      ExportTable.ExportClass.Add(IntToStr(temp));
      ExportTable.Super.Add(IntToStr(ReadINDEX(F)));
      BlockRead(F, temp, 4); ExportTable.ExportPackage.Add(IntToStr(temp));
      temp := ReadINDEX(F);
      ExportTable.ObjectName.Add(NameTable.ObjectName.Strings[temp]);
      BlockRead(F, temp, 4); ExportTable.ObjectFlags.Add(IntToStr(temp));
      ExportTable.SerialSize.Add(IntToStr(ReadINDEX(F)));
      if (StrToInt(ExportTable.SerialSize.Strings[i]) > 0) then
        ExportTable.SerialOffset.Add(IntToStr(ReadINDEX(F)))
      else
        ExportTable.SerialOffset.Add('-1');
    end;

    //Get object references (requires complete tables)
    for i := 0 to ImportTable.Count - 1 do
    begin
      ImportTable.ImportPackage.Strings[i] := GetObjectReference(NameTable, ImportTable, ExportTable, StrToInt(ImportTable.ImportPackage.Strings[i]));
    end;
    for i := 0 to ExportTable.Count - 1 do
    begin
      ExportTable.ExportClass.Strings[i] := GetObjectReference(NameTable, ImportTable, ExportTable, StrToInt(ExportTable.ExportClass.Strings[i]));
      ExportTable.Super.Strings[i] := GetObjectReference(NameTable, ImportTable, ExportTable, StrToInt(ExportTable.Super.Strings[i]));
      ExportTable.ExportPackage.Strings[i] := GetObjectReference(NameTable, ImportTable, ExportTable, StrToInt(ExportTable.ExportPackage.Strings[i]));
    end;

    //Create packagelist of internal packages + textures
    CreateGroupList;

    Result := TRUE;
  end;
end;

procedure TUTPackage.LoadNormalTexture;
begin
  //Free previous texture
  NormalTexture.Free;
  NormalTexture := TNormalTexture.Create;

  NormalTexture.Load(NameTable, ImportTable, ExportTable, PkgVer, Name, MaxPreviewSize, F);
end;

//Create a list of internal packages
procedure TUTPackage.CreateGroupList;
var
  i,k,j   : integer;
begin
  k := 0;
  TextureCount := 0;
  //get packages/textures from ExportTable
  for i:=0 to ExportTable.Count-1 do
  begin
    if (ExportTable.ExportClass.Strings[i] = 'Package') then
    begin
      Group[k].Name := ExportTable.ObjectName.Strings[i];
      for j:=0 to ExportTable.Count-1 do
        if (ExportTable.ExportClass.Strings[j] = 'Texture') then
          if  (ExportTable.ExportPackage.Strings[j] = Group[k].Name) then
              Group[k].Entries.Add(ExportTable.ObjectName.Strings[j]);
      k:=k+1;
    end;
  end;

  //Create [ALL] Group
  Group[k].Name := '[ALL]';
  for i:=0 to ExportTable.Count-1 do
  begin
    if (ExportTable.ExportClass.Strings[i] = 'Texture') then
    begin
      Group[k].Entries.Add(ExportTable.ObjectName.Strings[i]);
      TextureCount := TextureCount + 1;
    end;
  end;
  GroupCount := k+1;
end;

////////////////////////////////////////////////////////////////////////////////
//Save

procedure TUTPackage.Save;
var
  i,k         : integer;
  tempBYTE    : BYTE;
  tempDWORD   : DWORD;
  tempINT     : integer;
begin
  ////////////////////////////
  //Open file
  if ( IsPkgOpen = TRUE ) then
    raise TIllegalPackageOperationException.Create('UTSavePackage-object is already in use (TUTPackage.Save - IsPkgOpen = TRUE though it should be IsPkgOpen = FALSE)');
  AssignFile(F, FileName);
  ReWrite(F,1);
  IsPkgOpen := TRUE;

  ////////////////////////////
  //Set up progress-bar
  tempINT := 3 + GroupCount;
  for i:=0 to GroupCount-1 do
    tempINT := tempINT + AddGroup[i].TextureName.Count;
  MainForm.CreateUTXProgressBar.Visible := TRUE;
  MainForm.CreateUTXProgressBar.Min := 0;
  MainForm.CreateUTXProgressBar.Max := tempINT;
  MainForm.CreateUTXProgressBar.Position := 0;

  ////////////////////////////
  //Write Global Header
  Signature := $9E2A83C1;
  BlockWrite(F, Signature, 4);
  PkgVer := 69;
  BlockWrite(F, PkgVer, 4);
  PackageFlags := $0001;
  BlockWrite(F, PackageFlags, 4);
  //Table infos are written later (PackageFlags is only a dummy)
  BlockWrite(F, PackageFlags, 4);
  BlockWrite(F, PackageFlags, 4);
  BlockWrite(F, PackageFlags, 4);
  BlockWrite(F, PackageFlags, 4);
  BlockWrite(F, PackageFlags, 4);
  BlockWrite(F, PackageFlags, 4);
  //GUID and other stuff
  for i:=0 to 15 do
  begin
    tempBYTE := Random(255);
    BlockWrite(F, tempBYTE, 1);
  end;
  tempDWORD := 1;
  BlockWrite(F, tempDWORD, 4);
  BlockWrite(F, PackageFlags, 4); //Same as ExportCount
  BlockWrite(F, PackageFlags, 4); //Same as NameCount
  //Update ProgressBar
  MainForm.CreateUTXProgressBar.StepIt;

  //Create Standard Table-Entries
  //NameTable
  NameTable.ObjectName.Add('None');
  tempDWORD := $00010000 or $00020000 or $00040000 or $04000000 or $00000010 or $00000400;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('InternalTime');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('Color');
  tempDWORD := $00010000 or $00020000 or $00040000 or $04000000 or $00000010 or $00000400;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('MaxColor');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('MipZero');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('UBits');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('VBits');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('USize');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('VSize');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('Palette');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('UClamp');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('VClamp');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('Class');
  tempDWORD := $00010000 or $00020000 or $00040000 or $04000000 or $00000010 or $00000400;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('Engine');
  tempDWORD := $00010000 or $00020000 or $00040000 or $04000000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('Texture');
  tempDWORD := $00010000 or $00020000 or $00040000 or $04000000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('Core');
  tempDWORD := $00010000 or $00020000 or $00040000 or $04000000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('Package');
  tempDWORD := $00010000 or $00020000 or $00040000 or $04000000 or $00000010 or $00000400;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  NameTable.ObjectName.Add('bMasked');
  tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
  NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
  //ImportTable
  ImportTable.ClassPackage.Add('15');
  ImportTable.ClassName.Add('16');
  ImportTable.ImportPackage.Add('0');
  ImportTable.ObjectName.Add('15'); //Fi - Core (0)
  ImportTable.ClassPackage.Add('15');
  ImportTable.ClassName.Add('12');
  ImportTable.ImportPackage.Add('-4');
  ImportTable.ObjectName.Add('9');  //Fi - Palette (1)
  ImportTable.ClassPackage.Add('15');
  ImportTable.ClassName.Add('12');
  ImportTable.ImportPackage.Add('-4');
  ImportTable.ObjectName.Add('14'); //Fi - Texture (2)
  ImportTable.ClassPackage.Add('15');
  ImportTable.ClassName.Add('16');
  ImportTable.ImportPackage.Add('0');
  ImportTable.ObjectName.Add('13'); //Fi - Engine (3)
  ImportTable.ClassPackage.Add('15');
  ImportTable.ClassName.Add('12');
  ImportTable.ImportPackage.Add('-1');
  ImportTable.ObjectName.Add('16'); //Fi - Texture (4)
  //Update ProgressBar
  MainForm.CreateUTXProgressBar.StepIt;

  ////////////////////////////
  //Save textures and their palettes
  for i:=0 to GroupCount - 1 do
  begin
    //Create Table-Emtries for Group-Package//
    //NameTable
    NameTable.ObjectName.Add(AddGroup[i].GroupName);
    tempDWORD := $00010000 or $00020000 or $00040000 or $00000010;
    NameTable.ObjectFlags.Add(IntToStr(tempDWORD));
    //ExportTable
    ExportTable.ExportClass.Add('-5');
    ExportTable.Super.Add('0');
    ExportTable.ExportPackage.Add('0');
    ExportTable.ObjectName.Add(IntToStr(NameTable.ObjectName.Count-1));
    ExportTable.ObjectFlags.Add(IntToStr($00010000 or $00020000 or $00040000 or $00000004));
    //Package-object has only a None-Property (write Property and update Table)
    tempDWORD := 0;
    ExportTable.SerialSize.Add(IntToStr(SizeOfINDEX(tempDWORD)));
    ExportTable.SerialOffset.Add(IntToStr(FilePos(F)));
    WriteINDEX(tempDWORD, F);
    //Set PackageINDEX into ExportTable
    AddGroup[i].GroupIndex := ExportTable.ObjectName.Count-1;
    //Update ProgressBar
    MainForm.CreateUTXProgressBar.StepIt;

    //Save NormalTextures//
    for k:=0 to AddGroup[i].TextureName.Count-1 do
    begin
      NormalTexture.Save(NameTable, ImportTable, ExportTable, AddGroup[i], k, F);
      //Update ProgressBar
      MainForm.CreateUTXProgressBar.StepIt;
    end;
  end;

  ////////////////////////////
  //Save tables
  //NameTable
  NameTable.Count := NameTable.ObjectName.Count;
  NameTable.Offset:= FilePos(F);
  for i:=0 to NameTable.Count-1 do
  begin
    WriteName(NameTable.ObjectName.Strings[i], F);
    tempDWORD := StrToInt( NameTable.ObjectFlags.Strings[i]);
    BlockWrite(F, tempDWORD, 4);
  end;
  //ImportTable
  ImportTable.Count := ImportTable.ObjectName.Count;
  ImportTable.Offset:= FilePos(F);
  for i:=0 to ImportTable.Count-1 do
  begin
    WriteINDEX(StrToInt(ImportTable.ClassPackage.Strings[i]), F);
    WriteINDEX(StrToInt(ImportTable.ClassName.Strings[i]), F);
    tempDWORD := StrToInt(ImportTable.ImportPackage.Strings[i]);
    BlockWrite(F, tempDWORD, 4);
    WriteINDEX(StrToInt(ImportTable.ObjectName.Strings[i]), F);
  end;
  //ExportTable
  ExportTable.Count := ExportTable.ObjectName.Count;
  ExportTable.Offset:= FilePos(F);
  for i:=0 to ExportTable.Count-1 do
  begin
    tempDWORD := StrToInt(ExportTable.ExportClass.Strings[i]);
    WriteIndex(tempDWORD, F);
    Seek(F, FilePos(F)-1);
    tempDWORD := ReadINDEX(F);
    WriteIndex(StrToInt(ExportTable.Super.Strings[i]), F);
    tempDWORD := StrToInt(ExportTable.ExportPackage.Strings[i]);
    BlockWrite(F, tempDWORD, 4);
    tempDWORD := StrToInt(ExportTable.ObjectName.Strings[i]);
    WriteIndex(tempDWORD, F);
    tempDWORD := StrToInt(ExportTable.ObjectFlags.Strings[i]);
    BlockWrite(F, tempDWORD, 4);
    WriteIndex(StrToInt(ExportTable.SerialSize.Strings[i]), F);
    if ( StrToInt(ExportTable.SerialSize.Strings[i]) > 0 ) then
      WriteIndex(StrToInt(ExportTable.SerialOffset.Strings[i]), F);
  end;
  //Write Counts and Offsets
  Seek(F, 12);
  tempDWORD := NameTable.Count;
  BlockWrite(F, tempDWORD, 4);
  tempDWORD := NameTable.Offset;
  BlockWrite(F, tempDWORD, 4);
  tempDWORD := ExportTable.Count;
  BlockWrite(F, tempDWORD, 4);
  tempDWORD := ExportTable.Offset;
  BlockWrite(F, tempDWORD, 4);
  tempDWORD := ImportTable.Count;
  BlockWrite(F, tempDWORD, 4);
  tempDWORD := ImportTable.Offset;
  BlockWrite(F, tempDWORD, 4);
  //Update ProgressBar
  MainForm.CreateUTXProgressBar.StepIt;
  MainForm.CreateUTXProgressBar.Visible := FALSE;

  ////////////////////////////
  //Close file
  CloseFile(F);
  IsPkgOpen := FALSE;
end;

////////////////////////////////////////////////////////////////////////////////

end.
